open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Nivre;;
open Printers;;
open Surprisal;;
open Main;;

(*This is the main file of the software. It takes in command line arguments and runs the main argument*)

(*Command line argument handling*)
let evaluate () =
begin
  let output = ref "dependency" in
  let file = ref "" in
  let userK = ref 3 in
  let train = ref "WSJ" in
  let oneFileIn = ref "true" in
  let finalOnlyIn = ref "true" in
  let modelType = ref "conditional" in
  let grainType = ref "POS" in
  Arg.parse [ ("-file",Arg.Set_string file,"Input file in proper format (see README for specs).");
	      ("-output",Arg.Set_string output,"Output type (depPrint, dgGraph, dependency, or config). Default is dependency.");
	      ("-k",Arg.Set_int userK,"The k-value for k-best parsing.  Default is 3.");
	      ("-train",Arg.Set_string train,"Corpus-type for trained features (Negra, WSJ, Dutch, or ChTB). Default is WSJ.");
	      ("-oneFile",Arg.Set_string oneFileIn,"Option to print the output to one file (true or false). Default is true.");
              ("-finalOnly",Arg.Set_string finalOnlyIn,"Option to print only final dependencies or incremental dependencies (true or false). Default is true.");
             (* ("-modelType",Arg.Set_string modelType,"Model type (conditional or generative). Default is conditional.");*)
	     (* ("-grainType",Arg.Set_string grainType,"Granularity of conditioning elements (lexical or POS). Default is POS.");*)
	    ]
    (function s -> print_string ("didn't understand "^s^"\n"))
    "Invoke the Nivre incremental dependency parser";
  
  
  if !file = "" then failwith "input file required";

  let surpPrint = true in
  let finalOnly = (if (!finalOnlyIn = "true") then true else false) in
  let genCondModel = (if (!modelType = "generative") then (ref Generative) else (ref Conditional)) in
  let oneFile = (if ((compare !oneFileIn "true") == 0) then true else false) in 
  let grain = (if (!grainType = "POS") then (ref Category) else (ref Lexical)) in

(*Load input sentences using Lex and Yacc into a list called sentenceList*)
  let sentenceList = (print_string "Loading input sentences...\n");(flush stdout) in
  let sentenceList =
  begin
    Utilities.linenum := 1; 
    try DpYacc.sentenceList DpLex.token (Lexing.from_channel (open_in !file))
    with (Parsing.Parse_error ) ->
      (print_string ("Syntax error in line "^(string_of_int !Utilities.linenum)^"\n");
      [])
  end in
  
(* Load in the appropriate weight file*)
  let load1 = (print_string "Loading transition probabilities...\n"); (flush stdout) in
  let weights = 
    let corpus = (match (!train) with
       "Negra" -> "../Input/Weights/negraTiger"
       |"ChTB" -> "../Input/Weights/chinese"
       |"Dutch" -> "../Input/Weights/dutch"
       |_ -> "../Input/Weights/wsj") in
    match (!grain) with
	Lexical -> (corpus^"Lex.txt") 
      |_ ->  (corpus^"POSStacksNext.txt") in

    let runWeights = (inputWeights (open_in weights)) in  
    let rw = (print_string "Transition probabilities loaded...\n");(flush stdout) in

  let posArray = (Array.make 100 "") in
  let lexArray = (Array.make 100000 "") in

  let mapPOS, mapLex, featurePOS, posWord = 
    if (!genCondModel = Conditional) then ("","","","") else (
    let corpus = (match (!train) with
       "Negra" -> "../Input/Weights/negraTigerGen"
       |"ChTB" -> "../Input/Weights/chineseGen"
       |"Dutch" -> "../Input/Weights/dutchGen"
       |_ -> "../Input/Weights/wsjGen") in
    match (!grain) with
	Lexical -> (corpus^"POSMap.txt"), (corpus^"WordMap.txt"), (corpus^"LexPOS.txt"), (corpus^"POSWord.txt")
      |_ -> (corpus^"POSMap.txt"), (corpus^"WordMap.txt"), (corpus^"POSPOSStacksDIY.txt"), (corpus^"POSWordDIY.txt")) in
   
  let novelWords = findNovelWords sentenceList [] in 
  let nw = (print_string "Novel words found...\n");(flush stdout) in

(*Actually load the files*)
    let runGens = if (!genCondModel = Generative) then ((print_string "Loading lexical probabilities...\n");(flush stdout);(*(inputArray (open_in mapPOS) posArray); (inputWordHash (open_in mapLex) wordHash);*) (featureToPOS (open_in featurePOS)); (posToLexical (open_in posWord) novelWords lexicalizedHash)) else ""  in

(*The different feature models are specified here.  You have to add a new feature here*)

    let stack1Model = {list=[Stack1];name="Stack1"} in
    let stack2Model = {list=[Stack2];name="Stack2"} in 
    let stack3Model = {list=[Stack3];name="Stack3"} in 
    let stack1NextModel = {list=[Stack1Next];name="Stack1Next"} in 
    let stack2NextModel = {list=[Stack2Next];name="Stack2Next"} in 
    let stack3NextModel = {list=[Stack3Next];name="Stack3Next"} in
    let topModel = {list=[Top]; name="Top"} in 
    let nextModel = {list=[Next]; name="Next"}in 
    let distanceModel = {list = [Distance]; name="Distance"} in
    let positionModel = {list=[Position]; name="Position"} in
    let topChildrenModel = {list =[TopChildren]; name="TopChildren"} in
    let topParentModel = {list = [TopParent]; name="TopParent"} in
    let topLeftModel = {list=[TopLeft]; name="TopLeft"} in
    let topRightModel = {list =[TopRight]; name="TopRight"} in
    let stackSizeModel = {list=[StackSize];name="StackSize"} in
    let stack3TopCModel = {list=[Stack3;TopChildren];name="Stack3TopChildren"} in
    let allModel = {list=[Stack3;Stack1;Top;Next;Distance;Position;TopChildren;TopParent;TopLeft;TopRight;StackSize]; name="All"} in
    let stacksModel = {list=[Stack3;Stack2;Stack1];name="Stacks"} in
    let stacksNextModel = {list=[Stack3Next;Stack2Next;Stack1Next];name="StacksNext"} in

(*This list specifies the features that will actually be run.  Currently we're only running the stack features because they give us the best p and r*)
    let bestList = [stack3Model;stack3TopCModel;allModel;stack1Model] in 

    let stacks = [stacksModel;stack3Model;stack2Model;stack1Model] in
    let stacksNext = [stacksNextModel;stack3NextModel;stack2NextModel;stack1NextModel] in

(*Actually run the parser with runModel*)
    let rm = (print_string "Beginning parsing...\n"); (flush stdout); (runModel [stacksNextModel] sentenceList !output !userK surpPrint finalOnly oneFile) in

    if (rm == true) then (print_string ("Finished parsing "^(!file)^"\n")) else (print_string ("Something didn't work\n"))
end;;

evaluate();;
