open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Retrieval;;
open Nivre;;
open Printers;;
open PAndR;;
open Surprisal;;
open StateSpacePrinter;;


(** The parser driver *)

(** This is the module that acts as a hub, running the k-best search procedure and organizing both the input and output information. *)


(** {% \subsection{Preliminaries} %} *)
(**This dummyArc is just used as a default return type. *)
let dummyArc = {op=Shift; element=""; prob=0.0; fromWord={word="NULL";lemma="NULL";cat="NULL";pos=(-2);parent=(-2);toPeek=false};toWord={word="NULL";lemma="NULL";cat="NULL";pos=(-2);parent=(-2);toPeek=false};fprobProb=0.0;lexicalProb=1.0;peek=false};;


(*The item is the data structure that holds all of the parser state information (see utilities for the full spec). The set is used to hold different items during parsing. *)
module Item : FLOAT with type t = item = 
  struct 
    type t = item
    let  compare (x:t) (y:t) = if (x.forwardProb.fprob = y.forwardProb.fprob) then 0 else if (x.forwardProb.fprob > y.forwardProb.fprob) then 1 else -1
  end 
module ItemSet = Set.Make(Item) 

(** {% \subsection{Advance to next configuration} %} *)

(**This function actually makes the arcs that are specified by the probability model.
It cycles through the ArcSet, which is sorted from most probable to least, and returns an ItemSet, sorted from most probable Item to least.*)
let rec makeArcs k it arcs iSet  model=  if ((ArcSet.is_empty arcs) == true) then iSet else 
	 let topArc =  (ArcSet.max_elt arcs) in
	 let forward = (match topArc.fprobProb with
	     0.0 -> (it.forwardProb.fprob, (it.forwardProb.zeros+1))
	    |x -> (if (it.forwardProb.fprob = 0.) then (x, it.forwardProb.zeros) else ((it.forwardProb.fprob *. (x)), it.forwardProb.zeros))) in
	 match topArc.op with
	     Left -> (makeArcs (k+1) it (ArcSet.remove topArc arcs) (ItemSet.add {config=(leftArc it.config topArc.toWord.pos); arcInfo=topArc; history=it::it.history;rank=k;forwardProb={fprob=(fst forward);zeros=(snd forward); forSurprisalCalc=false;lexProb=topArc.lexicalProb}} iSet)  model)
	   |Reduce -> (makeArcs (k+1) it (ArcSet.remove topArc arcs) (ItemSet.add {config=(reduce it.config);arcInfo=topArc; history=it::it.history; rank=k;forwardProb={fprob=(fst forward);zeros=(snd forward); forSurprisalCalc=false;lexProb=topArc.lexicalProb}} iSet)  model)
	   |Right -> (makeArcs (k+1) it (ArcSet.remove topArc arcs) (ItemSet.add {config=(rightArc it.config topArc.fromWord.pos); arcInfo=topArc;  history=it::it.history; rank=k;forwardProb={fprob=(fst forward);zeros=(snd forward); forSurprisalCalc=true;lexProb=topArc.lexicalProb}} iSet)  model)
	   |Shift -> (makeArcs (k+1)it (ArcSet.remove topArc arcs)  (ItemSet.add{config=(shift it.config);arcInfo=topArc;history=it::it.history;rank=k;forwardProb={fprob=(fst forward);zeros=(snd forward); forSurprisalCalc=true;lexProb=topArc.lexicalProb}} iSet)  model);;


(** {% \subsection{K-best Search} %} *)

(** The next two methods are just safety-valves for the parse.  They make sure that there are no infinite loops, and if there are, they stop that parse. *)

(*
(**This function just makes sure there are no infinite loops through a parse of a sentence by checking the history and ensuring we're not taking the same action over and over.*)
let rec checkHistoryLoops hist arc = match hist with
    [] -> false
  |h::t -> if ((h.arcInfo.prob = arc.prob) && (h.arcInfo.toWord.pos = arc.toWord.pos))
    then (true)
    else (checkHistoryLoops t h.arcInfo);;

(**This function actually stops the loops.*)
let rec stopLoops arc hist = match hist with
    [] -> false
  |h::t -> if ((h.arcInfo.prob = arc.prob) && (h.arcInfo.toWord.pos = arc.toWord.pos))
    then (true)
    else (stopLoops arc t);;
*)


(**These next three functions run the k-best search*)


(**getChildren cycles through the current list of items (b) and finds all continuations (via mostProbable in probabilityModel.ml). It then creates the subsequent parser states via makeArcs above. The full set of items that continue off of b, sorted by most probable, are then returned as iSet*)
let rec getChildren b model iSet = match b with
         [] -> iSet
       |h::t -> (getChildren t model (makeArcs 1 h (mostProbable h.config model) iSet  model));;

(** This ranks the configurations from best to worst according to forward probability. *)
let rec bestC h c newC = match c with
	[] -> newC
  |j::tC -> if ((h.forwardProb.fprob) > (j.forwardProb.fprob)) then (bestC j tC (h::newC)) else (bestC h tC (j::newC));;


(**This finds the top k items returned from getChildren (i.e., the top k continuations)  and
adds the k-best to a list called c, which will become the current k-best in main. *)
let rec getKBest iSet k newC userK = match k with
    0 -> newC
  |_ -> match (ItemSet.is_empty iSet) with
	 true -> newC
       |false -> (let topItem = (ItemSet.max_elt iSet) in
		      let rankedItem = {topItem with rank=((userK-k)+1)} in
	  if ((List.length newC) > userK) then (getKBest (ItemSet.remove topItem iSet) (k-1) (newC@[rankedItem]) userK) else (getKBest (ItemSet.remove topItem iSet) (k-1) (newC@[rankedItem]) userK));;


(** {% \subsection{Are we there yet?} %} *)

(** These functions determine when to stop parsing. *)

(**This function cycles through the current k-best items and checks to see if any are final configurations. If so, they're separated from the rest and returned. So, items are returned in two lists, one that has final configs and the other that has configs that still need work*)
let rec finalConfig b finals rest = match b with
    [] -> (List.rev finals), (List.rev rest)
  |bh::bt -> 
     if (((List.length bh.config.stack) == 0) && ((List.length bh.config.tau) == 0))
    then (finalConfig bt (bh::finals) rest) 
    else (finalConfig bt finals (bh::rest));;

(**The main function that runs the search procedure.

This function has a stop (i) of 300 iterations as another prevention of loops. This is more than enough for any sentence in any of the corpora we have used in training.

It first checks to see if the current k-best list (b) has any final configs. If so, it adds them to a list called finals. For all non-final items, it runs getChildren and getKBest to get the next b. *)
let rec main b model holdEm i userK finals  = match i with
    300 -> (List.rev holdEm), finals
  |_ -> match (finalConfig b [] []) with 
	 [], [] -> (List.rev holdEm), finals
       |[], x -> (main (getKBest (getChildren x model ItemSet.empty ) userK [] userK) model (x::holdEm)  (i+1) userK finals )
       |y, [] -> (List.rev holdEm), (finals@y)
       |z, w -> (main (getKBest (getChildren w model ItemSet.empty ) userK [] userK) model (w::holdEm) (i+1) userK (finals@z) );;


(** {% \subsection{Input initialization} %} *)

(**This initializes the different elements that make up an item*)

(**Decay is used for retrieval.*)
let rec initializeDecay tau dl =  match tau with
    [] -> (List.rev dl)
  |h::t -> (initializeDecay t ({tNode=h;timeSince=[];fan=0.0}::dl));;

(**h and d are part of the Nivre configuration--they hold the analysis that's being built*)

let rec initializeH tau h = match tau with
    [] -> h
  |x::xs -> initializeH xs ((H.add (Int32.of_int x.pos) (-1) h))

let rec initializeD tau d = match tau with
    [] -> d
  |x::xs -> initializeD xs ((D.add (Int32.of_int x.pos) "right" d))


(** Initialize the parser state itself.*)
let initialize t = [{config = {stack=[]; tau=t; h=(initializeH t H.empty); d=(initializeD t D.empty); soFar=[]; time=0.0;decay=(initializeDecay t [])}; arcInfo=dummyArc;history = []; rank=1;forwardProb = {fprob=(0.0);zeros=0;forSurprisalCalc=false;lexProb=0.}}];;


(** {% \subsection{Find optimal parse.} %} *)
(**This function counts the number of attachments a finished analysis has, which is used to determine the best analysis for the sentence*)
let countAttachments h = 
  let hList = (outputH h) in
    (List.length hList);;

(**This function finds the best analysis on the basis of forward probability*)
let rec getFinal finals best = match finals with
 [] -> [best]
  |h::t -> match (h.forwardProb.fprob > best.forwardProb.fprob) with
true -> getFinal t h
      |false -> getFinal t best;;


(*let rec getFinal finals best = match finals with
    [] -> [best]
  |h::t -> match ((countAttachments h.config.h) > (countAttachments best.config.h)) with
	true -> getFinal t h
      |false -> match ((countAttachments h.config.h) = (countAttachments best.config.h)) with
	    false -> getFinal t best
	  |true ->  match(h.forwardProb.fprob > best.forwardProb.fprob) with
		true -> getFinal t h
	      |false -> getFinal t best;;*)


(** {% \subsection{Main drivers} %} *)
(**The next two functions run the whole software. runModel is run from argParse. It not only runs the parser, 
but also runs the printers and the psycholinguistic measures (surprisal and retrieval).*)


(** This function translates the user arguments into parser options, and then runs the parser for the sentences.  It cycles through the sentences, figures out what printers it needs, runs the parser, gets the output, and then prints it. *)
let rec run model sentenceList i out userK surpPrint finalOnly j onePrinter = match sentenceList with
    [] ->  []
  |h::t -> let printers = (match out with
        "tensor" -> (getPrinter model.name i)
	|"config" -> (getPrinter model.name i)
	|"kbest" -> (getPrinter model.name i)
      |"dgGraph" -> (getDGPrinters model.name i userK [])
      |"dependency" -> (onePrinter)
      |_ -> (getPrinter model.name i)) in 
           let he, finals = (main (initialize h) model [] 0 userK []) in 
	   let realFinal = (if ((List.length finals) > 0) then (getFinal finals (List.hd finals)) else []) in
	   let statePrinter,stockPrinter = (getStatePrinter model.name i),(getStockPrinter model.name i) in
	   let printStates = (statePrinterDriver he statePrinter stockPrinter [] userK) in
	   let surpArray = (if (surpPrint==true) then (surprisal he surprisalPrinter i) else (Array.make 1 0.0)) in
	   let actRPrint = (actRPrint he decayPrinter i userK) in 
	   let prints1 = (match finalOnly with
	       true -> (printIter [realFinal] printers out surpArray finalOnly)
	       |false -> (printIter he printers out surpArray finalOnly)) in
           let flushes = ((flush surprisalPrinter); (flush decayPrinter); (flushAll printers)) in
           let moreFlushes = (if ((compare out "dependency")==0) then (flushAll onePrinter) else (close_outAll printers)) in
           let outputWhere =  (match (i mod 5000) with
                  0 -> (print_string ((string_of_int i)^" sentences printed.\n"))
                 |_ -> match (i mod 500) with  
                       0 -> (print_string "*")   
                      |_ -> match (i mod 100) with
                            0 -> (print_string ".")
                           |_ -> (flush stdout)) in
             (realFinal::(run model t (i+1) out userK surpPrint finalOnly (j+1) onePrinter));;


(** This function creates the various printers, and cycles through them for different sentences.  Its input is a modelList, which means that it will cycle through all the sentences in the input file for each probability model (or feature) that the user wants to run.  The outputs are then feature-specific. *)
let runModel modelList sentenceList out userK surpPrint finalOnly oneFile = 
  let rec rm modelList sentenceList out userK surpPrint finalOnly oneFile i = match modelList with
    [] -> true
    |h::t -> let outputs = ((output_string surprisalPrinter ("***"^h.name^"\n"));
             (output_string decayPrinter ("***"^h.name^"\n"));
             (output_string actRPrinter ("***"^h.name^"\n"));
             (print_string ("\nCalculating "^h.name^"\n"))) in
             let onePrinter = ([getOnePrinter h.name]) in
             let run = (run h sentenceList 1 out userK surpPrint finalOnly 0 onePrinter) in
	     let pAndRPrint = (pAndRPrint run prPrinter h.name) in
             (rm t sentenceList out userK surpPrint finalOnly oneFile (i+1))
  in (rm modelList sentenceList out userK surpPrint finalOnly oneFile 0 );;

