open Utilities;;
open LoadFeatures;;

(**Find probabilities for transitions from a configuration. *)

(** This file runs the probability model for the software. It matches the test configuration with the probabilities that were obtained from the training data.  Note that this just gives the most probable arcs for a given configuration. The seach is handled by main.ml.*)


(** {% \subsection{Configuration properties} %} *)
(*Get Information about this Configuration*)

(**These functions just translate the current configuration into the particular features we use.*)

(** For the distance feature. *)
let distanceBetween s t = let i = s.pos in 
   let j = t.pos in (string_of_int (abs (i - j)))

(** For the topChildren feature. *)
let rec getTopChildren s h kids config = match h with
    [] -> if (kids = "") then "NULL" else kids
  |(a,b)::t -> try (if ((b == s.pos) == true) 
                   then (let word = (List.nth config.soFar (Int32.to_int a)) in
                         (getTopChildren s t (kids^"|"^(conditionOn word)) config))
                   else (getTopChildren s t kids config))
               with Invalid_argument "List.nth" -> (getTopChildren s t kids config)

(** For topParent.*)
let rec getTopParent s h c = match (conditionOn s) with
 "NULL"  -> "NULL"
|_ -> match h with
    [] -> "NULL"
  |(a,b)::t -> if ((Int32.compare (Int32.of_int s.pos) a) == 0) then 
	(try (conditionOn (List.nth c.soFar b)) 
	  with Invalid_argument "List.nth" -> (getTopParent s t c))
	else (getTopParent s t c);;


(** For TopLeft and topRight. *)
let rec getTopLeftorRight s h c = match h with
    [] -> "NULL"
  |(a,b)::t -> match ((List.length c.soFar) > (Int32.to_int a)) with
	true -> if ((b == s.pos) == true) then (try(conditionOn (List.nth c.soFar (Int32.to_int a))) with Invalid_argument "List.nth" -> (getTopLeftorRight s t c)) else (getTopLeftorRight s t c)
      |false -> "NULL";;

(** For Roots.*)
let rec getRoots current h c i = match h with
    [] -> (string_of_int i)
  |(a,b)::t -> let pos = (Int32.to_int a) in
      if (pos < current)
      then try (match (conditionOn (List.nth c.soFar pos)) with
		  "VBD"|"VBG"|"VBN"|"VBP"|"VBZ"|"VVFIN"|"VVIMP"|"VVINF"|"VVIZU"|"VVPP"|"VAFIN"|"VAIMP"|"VAINF"|"VAPP"|"VMFIN"|"VMINF"|"VMPP"|"VV"|"VA"|"VC"|"VE" -> 
		  if (b = (-1)) then (getRoots current t c (i+1)) else (getRoots current t c i)
		 |_ -> (getRoots current t c i)) with Invalid_argument "List.nth" -> (getRoots current t c i)
      else (string_of_int i);;

(**For Stack3Roots combined.*)
let rec getS3Roots s tau make n outH c = match n with
  0 -> make^"**"^(conditionOn tau)^"**"^(getRoots tau.pos outH c 0)
  |_ -> match s with
    [] -> make^"**"^(conditionOn tau)^"**"^(getRoots tau.pos outH c 0)
  |h::t -> getS3Roots t tau (make^(conditionOn h)^"|") (n-1) outH c;;


(**This is for stack features. Make nulls just adds any nulls required (so for Stack3, if the stack only has one element, it would add 2 nulls.*)
let rec makeNulls i s = match i with
  0 -> s
|_ -> (makeNulls (i-1) (s^"NULL|"));;

(** Makes the appropriate Stack+Input word feature (StacksNext). It takes in the stack (s), the next input word (tau) and aggregates feature informatin in (make). i determines how many elements of the stack you need--3 for Stack3, etc.*)
let rec makeStackNexts s tau make i = match i with
    0 -> make^"**"^(conditionOn tau)
  |_ -> match s with
    [] -> if (i > 0) then (make^(makeNulls i "")^"**"^(conditionOn tau)) else (make^"**"^(conditionOn tau))
  |h::t -> makeStackNexts t tau (make^(conditionOn h)^"|") (i-1);; 

(** This is similar to makeStackNexts except it does not include the next word.*)
let rec makeStacks s make i = match i with
    0 -> make
  |_ -> match s with
    [] -> if (i > 0) then (make^(makeNulls i "")) else (make)
  |h::t -> makeStacks t (make^(conditionOn h)^"|") (i-1);; 

(** {% \subsection{Generative probs} %} *)
(**This next section adds generative probability information to the discriminative probabilities. *)

(** Finds the generative probability associated with a POS.*)
let rec findPOSProb cat genPOS nullProb= match genPOS with
    [] -> nullProb
  |h::t -> if (h.genName = (String.uppercase cat)) then h.genProb else (if (h.genName = "NULL") then (findPOSProb cat t h.genProb) else (findPOSProb cat t nullProb));;

(** Finds the generative probability.*)
let generativeProb variable tau = match !genCondModel with
    Conditional -> 1.0
  |_ -> let key = ((String.uppercase tau.word)^"|"^(String.uppercase tau.cat)) in
       try (Hashtbl.find lexicalizedHash key) with Not_found -> ((print_string ("LexicalizedHash Not found "^key^"\n")); 1.0);; (*BACKOFF*)

(** {% \subsection{Finding the probabilities} %} *)

(**The probModel function aggregates the probabilitiy information for each transition possible from a configuration.  Because we often run with multiple features at once (for example, the Stacks features incorporates Stack3, Stack2, and Stack1), the probModel function multiplies the exponent of the weight for this feature by the probability that has already been collected. It then returns a new probability for the arc (actually a new arc with updated prob).*)
let rec probModel variable probList newList = match probList with
    [] -> (List.rev newList)
  |ph::pt -> let prob = (if (ph.prob = 0.) then 1. else ph.prob) in
      match ph.op with
	Left -> let newProb = (if (variable.left = 0.) then (ph.prob) else (prob *. (exp variable.left))) in
(probModel variable pt ({op=ph.op; element=(conditionOn ph.fromWord); prob=newProb; fromWord=ph.fromWord; toWord=ph.toWord;fprobProb=newProb;lexicalProb=ph.lexicalProb;peek=ph.peek}::newList) )
      |Right -> let newProb = (if (variable.right = 0.) then (ph.prob) else (prob *. (exp variable.right))) in
(probModel variable pt ({op=ph.op; element=(conditionOn ph.toWord); prob=newProb; fromWord=ph.fromWord; toWord=ph.toWord;fprobProb=newProb;lexicalProb=ph.lexicalProb *. (generativeProb variable ph.toWord);peek=ph.peek}::newList) )
      |Shift -> let newProb = (if (variable.shift = 0.) then (ph.prob) else (prob *. (exp variable.shift))) in
(probModel variable pt ({op=ph.op; element=(conditionOn ph.fromWord); prob=newProb; fromWord=ph.fromWord; toWord=ph.toWord;fprobProb=newProb;lexicalProb=ph.lexicalProb *. (generativeProb variable ph.toWord);peek=ph.peek}::newList))
      |Reduce -> let newProb = (if (variable.reduce = 0.) then (ph.prob ) else (prob *. (exp variable.reduce)))in
(probModel variable pt ({op=ph.op; element=(conditionOn ph.fromWord); prob=newProb; fromWord=ph.fromWord; toWord=ph.toWord;fprobProb=newProb;lexicalProb=ph.lexicalProb;peek=ph.peek}::newList));;


(**This function just finds the varProb element in the weights hash tables specific to the feature.*)
let findHashElement tbl str = try (Hashtbl.find tbl str)
                       with Not_found -> {left=0.0; right=0.0; shift=0.0; reduce=0.0; genPOS=[]};;

(**We were using this peekProb thing for generative probabilities.  Again, not crucial since this distribution does not use generative probabilities.*)
let rec getPeekProb cat pList = match pList with 
    [] -> 1.0
  |h::t -> if ((String.uppercase h.genName) = (String.uppercase cat)) then h.genProb else (getPeekProb cat t);;

let rec findPeekProb modelList config prob = match modelList with 
    [] -> (prob)
  |h::t ->  let hash,i = match h with
    Stack3Next -> stack3, 3
    |Stack2Next -> stack2, 2
    |_ -> stack1, 1 in
  let key = (makeStacks config.stack "" i) in
  let fhe = (findHashElement hash key) in
  let cat = try (List.hd config.tau).cat with Failure "hd" -> "NULL" in
  (findPeekProb t config (prob +.(getPeekProb cat fhe.genPOS)));;

(**This function acts as a kind of driver to find the appropriate probabilities specific to the feature model that is being run.  It cycles through the list of features that we would like to run (e.g. Stack1, Stack2, Stack3), gets the probabilities for each of the transitions (Left, Right, Shift, Reduce), and then adds them to this probList item, which aggregates the probabilities across features.

This function takes in:
- a config (c), 
- a list of models (i.e., Stack3 + Stack2 + Stack1) 
- and a list of arcs (probList), and cycles through the models. 

It matches the model to the appropriate feature key and hash table, and then sends the weight information off to probModel to recalculate the probabilities for each of the arcs in the model. It returns the list of arcs (probList), which have the updated weights.*)
let rec matchModel c modelList probList = match modelList with
    [] -> probList
  |hm::tm -> let s = if (((List.length c.stack) > 0) == true) then (List.hd c.stack) else ({word="NULL";cat="NULL";lemma="NULL";pos=(-2);parent=(-2);toPeek=false}) in
             let t = if (((List.length c.tau) > 0) == true) then (List.hd c.tau) else ({word="NULL";cat="NULL";lemma="NULL";pos=(-2);parent=(-2);toPeek=false}) in
	     let peek = try (List.hd probList).peek with Failure "hd" -> false in
             (match hm with
                Top -> (matchModel c tm (probModel (findHashElement top (conditionOn s))  probList []))
               |Next -> (matchModel c tm (probModel (findHashElement next (conditionOn t)) probList []))
               |Stack1Next -> (matchModel c tm (probModel (findHashElement stack1Next (makeStackNexts c.stack t "" 1)) probList []))
               |Stack2Next -> (matchModel c tm (probModel (findHashElement stack2Next (makeStackNexts c.stack t "" 2)) probList []))
               |Stack3Next -> (matchModel c tm (probModel (findHashElement stack3Next (makeStackNexts c.stack t "" 3)) probList []))
               |Stack1 -> (matchModel c tm (probModel (findHashElement stack1 (makeStacks c.stack "" 1)) probList []))
               |Stack2 -> (matchModel c tm (probModel (findHashElement stack2 (makeStacks c.stack "" 2)) probList []))
               |Stack3 -> (matchModel c tm (probModel (findHashElement stack3 (makeStacks c.stack "" 3)) probList []))
               |Distance -> (matchModel c tm (probModel (findHashElement distance (distanceBetween s t)) probList []))
               |Position -> (matchModel c tm (probModel (findHashElement position (string_of_int t.pos))  probList []))
               |TopChildren -> (matchModel c tm (probModel (findHashElement topChildren (getTopChildren s (outputH (c.h)) "" c))  probList []))
               |TopParent -> (matchModel c tm (probModel (findHashElement topParent (getTopParent s (outputH c.h) c)) probList []))
               |TopLeft -> (matchModel c tm (probModel (findHashElement topLeft (getTopLeftorRight s (outputH (c.h)) c)) probList []))
               |StackSize -> (matchModel c tm (probModel (findHashElement stackSize (string_of_int (List.length c.stack))) probList []))
               |TopRight -> (matchModel c tm (probModel (findHashElement topRight (getTopLeftorRight s (List.rev (outputH (c.h))) c)) probList [])));;


(** This function creates the empty (i.e., no probs)  arcs (probList) that can be taken from this configuration. Notice that you can only do a left arc if the current input word doesn't have a parent.*) 
let makeProbsList c peek =
  let s = if (((List.length c.stack) > 0) == true) then (List.hd c.stack) else ({word="NULL";lemma="NULL";cat="NULL";pos=(-2);parent=(-2);toPeek=false}) in
  let t = if (((List.length c.tau) > 0) == true) then (List.hd c.tau) else ({word="NULL";lemma="NULL";cat="NULL";pos=(-2);parent=(-2);toPeek=false}) in
    match (List.length c.stack) with
      0 -> [{op=Shift; element="";prob=0.0; fromWord=t; toWord=t; fprobProb=0.0; lexicalProb=1.;peek=peek}]
     |_ -> match (List.length c.tau) with 
	   0 -> [{op=Reduce; element=""; prob=0.0; fromWord=s; toWord=s;fprobProb=0.0;lexicalProb=1.;peek=peek}]
	 |_ -> match (getTopParent s (outputH c.h) c) with
		"NULL" -> [ {op=Left; element=""; prob=0.0; fromWord=t; toWord=s;fprobProb=0.0;lexicalProb=1.0;peek=peek};
			    {op=Shift; element=""; prob=0.0; fromWord=t; toWord=t;fprobProb=0.0;lexicalProb=1.0;peek=peek};
			    {op=Right; element=""; prob=0.0; fromWord=s; toWord=t;fprobProb=0.0;lexicalProb=1.0;peek=peek};
			    {op=Reduce; element=""; prob=0.0; fromWord=s; toWord=s;fprobProb=0.0;lexicalProb=1.0;peek=peek}
			  ]
	      |_ -> [ {op=Shift; element=""; prob=0.0; fromWord=t; toWord=t;fprobProb=0.0;lexicalProb=1.0;peek=peek};
		      {op=Reduce; element=""; prob=0.0; fromWord=s; toWord=s;fprobProb=0.0;lexicalProb=1.0;peek=peek};
		      {op=Right; element=""; prob=0.0; fromWord=s; toWord=t;fprobProb=0.0;lexicalProb=1.0;peek=peek}
		    ];;

(** This creates a set of arcs from the accumulated arc probabilities in probList. The arcSet is then automatically ordered from high probability to low. See utilities for specs.*)
let rec getArcs arcs probList = match probList with 
	[] -> arcs
      |h::t -> match ((h.op == Left) || (h.op == Right)) with
	    true -> (match (((compare h.fromWord.word "NULL")==0) || ((compare h.toWord.word "NULL") == 0)) with
		true -> (getArcs arcs t)
	      |false -> (getArcs (ArcSet.add h arcs) t))
	    |false -> (getArcs (ArcSet.add h arcs) t);;

(** The following functions are used to normalize the weights that are currently in probList and turn them into probabilities. *)

(** This gets the total among all the 4 weights.*)
(*No longer needed--now done with awk scripts after learning*)
let rec totalProbs pl tot = match pl with
    [] -> tot
  |h::t -> totalProbs t (tot+. h.prob);;

(**And this normalizes by dividing each exp weight by its total. Note that the weights were already exponent-ed in probModel above. *)
let normalizeProbs pl = 
  let total = (totalProbs pl 0.0) in
  if (total <= 0.) then pl else (
  let rec normalize pl tot = match pl with
      [] -> []
    |h::t -> {op=h.op;element=h.element;prob=((h.prob/.tot));fromWord=h.fromWord;toWord=h.toWord;fprobProb=((h.fprobProb/.tot));lexicalProb=h.lexicalProb;peek=h.peek}::(normalize t tot)
  in (normalize pl total));; 

(**More peeking, for generative probs.*)
let peekProbs config ml pl = 
  if ((List.hd pl).peek = false) then pl else (
  let peekProb = (findPeekProb ml config 0.0) in 
  let avgPeekProb = (peekProb /. (float_of_int (List.length ml))) in
  if (peekProb <= 0.) then  pl else (
    let rec peek avgPeekProb pl = match pl with
	[] -> []
      |h::t ->{op=h.op;element=h.element;prob=((h.prob*.avgPeekProb));fromWord=h.fromWord;toWord=h.toWord;fprobProb=((h.fprobProb*.avgPeekProb));lexicalProb=h.lexicalProb;peek=h.peek}::(peek avgPeekProb t) in
      (peek avgPeekProb pl)));;

(**Peeking.*)
let checkPeek config = 
  if ((List.length config.tau) = 0) then false, config else (
  let t = (List.hd config.tau) in
  let peek = t.toPeek in
  let newT = {word=t.word;lemma=t.lemma;cat=t.cat;pos=t.pos;parent=t.parent;toPeek=false} in
  let newTau = newT::(List.tl config.tau) in
  if peek then (peek, {stack=config.stack;tau=newTau;h=config.h;d=config.d;soFar=config.soFar;time=config.time;decay=config.decay}) else (peek, config));;

(**This function is called by getChildren in main.ml. It takes in a current config and a feature model (like the Stacks feature) and returns a set of possible arcs from the config ordered by their probabilities.*)
let mostProbable config model = match config.tau, config.stack with
    [],[] -> (ArcSet.empty)
  |_ -> 
     let peek, newConfig = (checkPeek config) in
       (getArcs ArcSet.empty (peekProbs newConfig model.list (normalizeProbs (matchModel config model.list (makeProbsList newConfig peek)))));;


