open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;

(**Retrieval Calculation.*)

(**This aggregates the functions used to calculate retrieval.*)

(** {% \subsection{Preliminaries} %} *)

(**Currently productions take 50ms., reading time takes 1ms, as specified by Lewis and Vasishth 2005*)
let production = 50.;;
let reading = 1.;;


(**This is similarity-based interference by POS.
NOTE: CURRENTLY ONLY FOR NEGRATIGER.  *)
let similarCats = [["ADJA";"ADJD"];["ADV";"PROAV"];["APPR";"APPRART";"APPO";"APZR"];["ART";"CARD"];["KOUI";"KOUS";"KON";"KOKOM"];["NN";"NE";"NNE"];["PDS";"PDAT"];["PIS";"PIAT"];["PPOSS";"PPOSAT"];["PRF";"PPER"];["PRELS";"PRELAT"];["PWS";"PWAT";"PWAV"];["PTKZU";"PTKNEG";"PTKVZ";"PTKANT";"PTKA"];["VVFIN";"VVIMP";"VVINF";"VVIZU";"VVPP";"VAFIN";"VAIMP";"VAINF";"VAPP";"VMFIN";"VMINF";"VMPP";"XY"];["$,";"$.";"$("]];;

(** {% \subsection{Similarity-based interference} %} *)
(**The following functions are all used for similarity-based interference.*)

(** findWord finds the word in the decayList (dl) that has position (pos).*)
let rec findWord dl pos = match dl with
    [] -> {tNode={word="";lemma="";cat="";pos=(-2);parent=(-2);toPeek=false};timeSince=[];fan=0.}
  |h::t -> if (h.tNode.pos == pos) then h else (findWord t pos);;

(** This function finds the sum of the time since the last retrieval, where tl is the list of times the word was retrieved, time is the current time, and sum aggregates.*)
let rec findSumTimeSince tl time sum = match tl with
    [] -> sum
  |h::t -> findSumTimeSince t time (sum +. ((time -. h)**(-0.5)));;

(**This function takes in a category (POS) and the similarCats list and returns the similarCats list specific (i.e., nouns, verbs, etc.*)
let rec fcs cat list fullList = match list with
    [] -> []
  |h::t -> if (cat = h) then (fullList) else (fcs cat t fullList);;
  
(*This function checks to see if the other word considered that we have parsed has a similar category to this word. *)
let rec findSimilarCats cat scs = match scs with
    [] -> []
  |h::t -> let list = (fcs cat h h) in
      if (List.length list = 0) then (findSimilarCats cat t) else (list);;

let rec similar sl cat = match sl with
    [] -> false
  |h::t -> if (cat = h) then true else (similar t cat)

(** Interference checks to see if the current word interferes with the retrieved word. *)
let interference thisWord retrievedWord =
  if (thisWord.tNode.cat = retrievedWord.tNode.cat) then (true) else (
    let similarList = (findSimilarCats retrievedWord.tNode.cat similarCats) in
     (similar similarList thisWord.tNode.cat));; 

(** Calculates weighted Sji. *)
let rec findSji w soFar decay sum retrieved = match soFar with 
    [] -> sum
  |h::t -> if (h.pos > -1) then 
      (let decayItem = (findWord decay h.pos) in
       let interferes = (interference decayItem retrieved) in
       if (interferes) then
       (let newSum = (w *. (1.5 -. log decayItem.fan)) in
	 findSji w t decay (sum+.newSum) retrieved)
       else findSji w t decay sum retrieved)
    else findSji w t decay sum retrieved;;

(** Finds number of words so far that interfere. *)
let rec findDistractors soFar decay retrieved number = match soFar with
[] -> number
|h::t -> if (h.pos < 0) then (findDistractors t decay retrieved number) else(
         let decayItem = (findWord decay h.pos) in
         if (interference decayItem retrieved) 
         then (findDistractors t decay retrieved (number+1))
         else (findDistractors t decay retrieved number));;

(** {% \subsection{Activation calculation} %} *)

(** Calculates activation. *)
let findActivation config i = 
  let decayItem = (findWord config.decay i) in 
  let sumTimeSince = (findSumTimeSince decayItem.timeSince (config.time +. production +. reading) 0.0) in
  let baseline = (log sumTimeSince) in
  let fan = ((findDistractors config.soFar config.decay decayItem 0)+1) in
  let w = 1.0 in
  let wSji  = (w *. (1.5 -. (log (float_of_int fan)))) in
let activation = baseline +. wSji in
  let e = exp (0. -. activation) in
  let latency = ((0.14*.e)*.100.) in
 latency;;

(**This function is used by nivre.ml to create a new decayList that is updated for the next configuration. *)
let rec updateDecay decayList i time newL = match decayList with
     [] -> (List.rev newL)
  |h::t -> 
     if (h.tNode.pos == i) 
     then 
      (let newDecay = {tNode = h.tNode; timeSince=(time::h.timeSince); fan=(h.fan+.1.0)} in
	(updateDecay t i time (newDecay::newL)))
     else (updateDecay t i time (h::newL));;

(** Finds the children of a particular word. *)
let rec findAllChildren h i kids = match h with
    [] -> kids
  |(k,e)::t -> if (e = i) then (findAllChildren t i ((Int32.to_int k)::kids)) else (findAllChildren t i kids);;

(** Finds the activation for a given word. *)
let rec findAllActivation config l act = match l with
    [] -> act
  |h::t -> (findAllActivation config t (act +. (findActivation config h)));; 

(** This function is used directly by nivre.ml to find the cost of a particular action. *)
let findTimeCost config action i = match action with
  Shift -> (let newTime = (config.time +. production +. reading) in 
	       (newTime, (updateDecay config.decay i newTime [])))
  |Reduce -> (let newTime = (config.time +. production) in
               (newTime, (updateDecay config.decay i newTime [])))
  |Left -> ((*let subDependents = (findAllChildren (outputH config.h) i []) in
	    let activation = (findAllActivation config (i::subDependents) 0.)*)
      let activation = (findActivation config i) in
      let newTime = (config.time +. production +. reading  +. activation +. production) in
	(newTime, (updateDecay config.decay i newTime [])))
  |Right ->  (let activation = (findActivation config i) in
	      let newTime = (config.time +. production +. reading +. activation +. production) in 
	      let newDecay = (updateDecay config.decay i newTime []) in 
		(newTime, (updateDecay newDecay (List.hd config.tau).pos newTime [])));;

(** {% \subsection{Printing out retrieval information} %} *)
(** There are two retrieval printers coded.  One prints out much like the ACT-R output given in Vasishth & Lewis 2005.  The other prints out in the format of surprisal for the parser (sentence tab word tab totalTime tab thisWordtime).  *)

(** {% \subsubsection{ACT-R printer} %} *)

(** Output printers. *)
let decayPrinter = (open_out "../actRTime.txt");;
let actRPrinter = (open_out "../timePredictions.txt");;

(** ms to seconds. *)
let convertTime t = (t /. 1000.);;


(** Formatting/static output. *)
let finalPrint t p = (output_string p ("TOTAL TIME FOR SENTENCE: "^(string_of_float (convertTime t))^" seconds.\n\n"));;

(** Formatting/static output. *)
let decisionTime word p time = (output_string p ("******************************************** \n
Action Time for \""^word^"\":  "^(string_of_float (convertTime time))^"\n******************************************** \n\n\n")); (flush p);;


(** Formatting/static output.*)
let readingPrint word p time= try (output_string p ("READING: \""^word^"\" at "^(string_of_float (convertTime time))^"\n\n")) with Failure "hd" -> (flush p); (flush p);;

(** Print out shift actions. *)
let shiftPrint item p prevTime array = 
  let decision = (item.config.time -. prevTime) in
  let wordPos = ((List.hd item.config.stack).pos) in
  let arrayItem = array.(wordPos) in
  let output = ((output_string p ((string_of_float (convertTime item.config.time))^": SHIFT \""^(List.hd item.config.stack).word^"\".\n\n")); (decisionTime ((List.hd item.config.stack).word) p decision); (flush p)) in
  let newArray = (Array.set array wordPos (decision::arrayItem)) in
   array;;

(** Print out the reduce actions. *)
let reducePrint item p prevTime array= 
  let decision = (item.config.time -. prevTime) in
  let wordPos = item.arcInfo.toWord.pos in
  let arrayItem = array.(wordPos) in
  let newArray = (Array.set array wordPos (decision::arrayItem)) in
    (output_string p ((string_of_float (convertTime item.config.time))^": REDUCE \""^item.arcInfo.toWord.word^"\".\n\n")); (decisionTime item.arcInfo.toWord.word p decision);(flush p); array;;

(** Print out the leftArc actions. *)
let leftPrint item p prevTime array= 
  let decision = (item.config.time -. prevTime) in
  let wordPos = item.arcInfo.fromWord.pos in
  let arrayItem = array.(wordPos) in
  let newArray = (Array.set array wordPos (decision::arrayItem)) in
  let activation = (item.config.time -. prevTime -. production -. production -. reading) in
    (output_string p ((string_of_float (convertTime (prevTime +. production)))^": Decide to attach \""^item.arcInfo.toWord.word^"\" as dependent.\n"));
    (output_string p ((string_of_float (convertTime (prevTime +. production +. activation)))^": Retrieve \""^item.arcInfo.toWord.word^"\".\n"));
    (output_string p ((string_of_float (convertTime (item.config.time)))^": LEFT-ARC with \""^item.arcInfo.toWord.word^"\".\n")); 
    (decisionTime item.arcInfo.fromWord.word p decision);
    (flush p); array;;

(** Print out the rightArc actions. *)
let rightPrint item p prevTime array =
  let decision = (item.config.time -. prevTime) in
  let wordPos = item.arcInfo.toWord.pos in
  let arrayItem = array.(wordPos) in
  let newArray = (Array.set array wordPos (decision::arrayItem)) in
  let activation = (item.config.time -. prevTime -. production -. production -. reading) in
    (output_string p ((string_of_float (convertTime (prevTime +. production)))^": Decide to attach \""^item.arcInfo.fromWord.word^"\" as parent.\n"));
    (output_string p ((string_of_float (convertTime (prevTime +. reading +. activation +. production)))^": Retrieve \""^item.arcInfo.fromWord.word^"\".\n"));
    (output_string p ((string_of_float (convertTime item.config.time))^": RIGHT-ARC with \""^item.arcInfo.fromWord.word^"\".\n")); 
    (decisionTime item.arcInfo.toWord.word p decision);
    (flush p);array;;

(** This organizes actions by word.  Shift and RightArc advance to the next word, whereas LeftArc and Reduce do not. *)
let rPrint h p prevTime prevAction first = match prevAction with
    Shift -> if (first) then (flush p) else (match h.arcInfo.op with
	Right -> (readingPrint h.arcInfo.toWord.word p prevTime)
      |_ -> (readingPrint h.arcInfo.fromWord.word p prevTime))
  |Right -> (readingPrint h.arcInfo.toWord.word p prevTime)
  |Left|Reduce -> (flush p);;

(** This outputs the sentences retrievals in ACT-R format.  It cycles through the history of the sentence. *) 
let rec printDecayHistory histList p prevTime prevAction first array = match histList with
    [] ->  (flush p); array
  |h::t -> if ((List.length h.config.tau == 0) && h.arcInfo.op = Reduce) then ((finalPrint h.config.time p);(flush p); array) else
    let rPrint = (rPrint h p prevTime prevAction first) in
    let newArray =  (match h.arcInfo.op with
	Shift -> (shiftPrint h p prevTime array)
      |Reduce -> (reducePrint h p prevTime array)
      |Left -> (leftPrint h p prevTime array)
      |Right -> (rightPrint h p prevTime array)) in
	  (printDecayHistory t p h.config.time  h.arcInfo.op false newArray);;

(** This method is called from main.ml to print out the ACT-R format.  Currently ACT-R only works with the winning parse, not for k-best or un-winning parses.  That is why I first get the winner and then cycle through its history.  *)
let printWinners fcl p =
  let winner = (List.hd (List.nth fcl ((List.length fcl) -1))) in
  let history = (List.rev winner.history) in
        let rp = (try (readingPrint (List.hd (List.hd history).config.tau).word p (List.hd history).config.time) with Failure "hd" -> (flush p)) in
	let array = (Array.make (List.length (List.hd history).config.tau) [])in
	(try (printDecayHistory (List.tl history) p 0.0 Shift true array)  with Failure "tl" -> array)


(** {% \subsubsection{Just retrieval times printing} %} *)

(*TO PRINT THE RETRIEVAL FOR ALL K'S*)

(** This aggregates the retrievals across a parse into an array of words.  That way all retrievals/actions that are associated with a single word are in a cell of an array.*)
let makeInput fcl = match fcl with
    [] -> Array.make 1 []
  |h::t -> match h with
	[] -> Array.make 1 []
      |hh::tt -> (Array.make ((List.length hh.config.tau)) []);;

(** Gets a total parse time for a word. *)
let rec calcTimePerWord l a = match l with
    [] -> a
  |h::t -> calcTimePerWord t (a +. h);;

(** This just prints the times, in the surprisal format.*)
let rec printTimesOnly sent array i fullTime= 
  if (i >= (Array.length array)) 
  then (flush actRPrinter) 
  else (let timeList = array.(i) in
	let timeForWord = (calcTimePerWord timeList 0.0) in
	let newTotal = (fullTime +. timeForWord) in
	let output = (output_string actRPrinter ((string_of_int sent)^"\t"^(string_of_int i)^"\t"^(string_of_float newTotal)^"\t"^(string_of_float timeForWord)^"\n")) in
	  (printTimesOnly sent array (i+1) newTotal));;

(** This ensures that the retrievals are computed by word, not by action. *)
let findRetrievalPos it = match it.config.tau with
    	[] -> (match it.arcInfo.op with 
	    Shift|Right ->(match it.config.stack with
		[] -> -2
	      |hh::tt -> ((List.hd it.config.stack).pos))
	  |_ -> -2)
       |h::t -> (match it.arcInfo.op with
	   Left|Reduce -> ((List.hd it.config.tau).pos)
	   |Shift|Right -> (match it.config.stack with
		 [] -> -2
	       |hhh::ttt ->((List.hd it.config.stack).pos)));;

(** This returns the parsing time before the word is parsed. *)
let rec findInitialWordTime hist = match hist with
    [] -> 0.0
  |h::t -> match h.arcInfo.op with
	Shift|Right -> h.config.time
      |Left|Reduce -> (findInitialWordTime t);;

(*let rec addNumbers kList array = match kList with
    [] -> ()
  |h::t -> match h.arcInfo.op with
	Left|Reduce -> (addNumbers t array)
      |Shift|Right -> 
	 let pos = (findRetrievalPos h) in
	   if (pos < 0) then (addNumbers t array) else (
	    let firstTime = (findInitialWordTime h.history) in
	     let thisTime = (h.config.time) in
	     let decision = (thisTime -. firstTime) in
	     let arrayItem = array.(pos) in
	     let na = (array.(pos) <- (decision::arrayItem)) in
	       (addNumbers t array));;*)

(** This aggregates the retrieval times for a word. *)
let rec addNumbers kList array = match kList with
   [] -> ()
  |h::t -> if (h.forwardProb.forSurprisalCalc) then (
      let pos = (findRetrievalPos h) in
      if (pos < 0) then (addNumbers t array) else (
          let firstTime = (findInitialWordTime h.history) in
          let thisTime = (h.config.time) in
          let decision = (thisTime -. firstTime) in
          let arrayItem = array.(pos) in
          let na = (array.(pos) <- (decision::arrayItem)) in
            (addNumbers t array)))
else (addNumbers t array);;

(** If we're looking at the mean retrieval for each word, this just takes all of the k retrieval times for a word and divides by k. *)
let rec meanRetrieval array k i = if (i = (Array.length array)) then (array) else
     (let element = array.(i) in
     let totalTime = (calcTimePerWord element 0.0) in
     let divideByK = (totalTime /. (float_of_int (List.length element)))in
     let update = array.(i) <- [divideByK] in
       (meanRetrieval array k (i+1)));;

(**Alternatively, we can use the maximum retrieval time for a word across k's.  This is what is used in the Boston et al. 2009 study. *)
let rec maxTimePerWord l time = match l with
    [] -> time
  |h::t -> if h > time then (maxTimePerWord t h) else (maxTimePerWord t time);;

(** Finds the maximum retrieval time. *)
let rec maxRetrieval array i = if (i = (Array.length array)) then (array) else
  (let element = array.(i) in
   let maxTime = (maxTimePerWord element 0.0) in
   let update = array.(i) <- [maxTime] in
     (maxRetrieval array (i+1)));;

let rec runNumbers fcl array= match fcl with
    [] -> ()
  |h::t -> (addNumbers h array); (runNumbers t array);;


(** The driver for this printer.  It takes in the configurations for a parser and returns the retrieval numbers per word. *)
let actRPrint fullConfigList p i userK = match (List.length fullConfigList) with
   0 -> flush p
|_ ->
  let output = (output_string p ("\n\nSENTENCE  "^(string_of_int i)^"\n")) in
  let array = (printWinners fullConfigList p) in
  let newArray = (makeInput fullConfigList) in
  let rn = (runNumbers fullConfigList newArray) in
  (*let mean = (meanRetrieval newArray userK 0) in*)
  let max = (maxRetrieval newArray 0) in
  (printTimesOnly i newArray 0 0.0);
  (output_string actRPrinter "\n"); 
  (flush p); (flush actRPrinter);;
