open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Nivre;;
open Printers;;


(** Surprisal calculator and printer. *)
(** It's important to remember that surprisal is just the log_2 of the prefix probability of the last word divided by the prefix probability of the current word. Or:

log_2 ((PrefixProb(word_i-1) / (PrefixProb(word_i))

The prefix probability is just the sum of the forward probabilities at a particular word across different k's.  The forward probability of a parse is kept inside the item, so each configuration has a forward probability.  Of these, only certain ones would be useful for a surprisal calculation (some parses do not advance to the next word, so it's crucial that we only calculate prefix probability based on the last transition at a particular word).  

So, once we have the crucial forward probabilities (which will be marked with a TRUE in the forSurprisalCalc field), we just add them for each k, and get the prefix probability for a word.  We store these prefix probabilities per word in an array called prefixProb. And then we use this to calculate surprisal. *)

(** The printer itself. *)
let surprisalPrinter = (open_out "../surprisal.txt");;

(** {% \subsection{Getting prefix probabilities} %} *)

(** This function updates the prefix probability in the prefix probability array (which is organized by word and aggregates prefix probabilities for different transitions at each word) based on the input word, i.*)
let rec checkH h i = match h with
    [] -> i
  |(k,e)::t -> if (e == i) then (i+1) else (checkH t i);;

(**This ensures that the prefix probabilities are computed by word, not by action. *)
let findSurpPosition it = match it.config.tau with
    	[] -> (match it.arcInfo.op with 
	    Shift|Right ->(match it.config.stack with
		[] -> -2
	      |hh::tt -> ((List.hd it.config.stack).pos))
	  |_ -> -2)
       |h::t -> (match it.arcInfo.op with
	   Left|Reduce -> ((List.hd it.config.tau).pos)
	   |Shift|Right -> (match it.config.stack with
		 [] -> -2
	       |hhh::ttt ->((List.hd it.config.stack).pos)));;

(**This goes through each k and calculates the surprisal at that word at each k.*)
let rec goThroughK k pa = match k with
    [] -> pa 
  |h::t -> let position = (findSurpPosition h) in
	      if (position > -1) 
	      then (match h.forwardProb.forSurprisalCalc with
			true -> (let newVal = (((h.forwardProb.fprob)*. h.forwardProb.lexProb) +. (Array.get pa position)) in
		 ((Array.set pa position newVal); (goThroughK t pa)))
		      |false -> (goThroughK t pa))
	      else (goThroughK t pa);;
     
(**This outputs the prefixprob array.*)
let rec outputArray pa i = match ((Array.length pa) == (i+1)) with
    true -> (flush surprisalPrinter)
  |false -> (outputArray pa (i+1));;

(**Outputs the prefix probs. *)
let rec getPPs fcl pa = match fcl with
    [] -> (outputArray pa 0); pa
  |h::t -> (getPPs t (goThroughK h pa)) ;;

(**Makes the empty array to store prefix probabilities. *)
let makeInput fcl = match fcl with
    [] -> Array.make 1 0.0 
  |h::t -> match h with
	[] -> Array.make 1 0.0
      |hh::tt -> (getPPs fcl (Array.make ((List.length hh.config.tau)) 0.0));;
    

(** {% \subsection{Calculating and printing surprisal} %} *)
(** Calculates surprisal based on the prefix probability array (by word) ppa, and returns a surprisal calc array (sca). i is used to ensure we have the right number of words. *) 
let rec calcSurp ppa sca i = match i with
    0 ->
      (Array.set sca i (abs_float ((log ( (1.0) /.( (Array.get ppa i))))/.(log 2.0))));sca
  |_ -> 
       (Array.set sca i (abs_float ((log (( (Array.get ppa (i-1)))/.( (Array.get ppa i))))/.(log 2.0))));
       (calcSurp ppa sca (i-1));;

(** Prints surprisal. *)
let rec printSurprisal tau p i j ppa array = match (j < (Array.length array)) with
    true -> ((output_string p ((string_of_int i)^"\t"^(string_of_int (j))^"\t"^(string_of_float (Array.get ppa j))^"\t"^(string_of_float (Array.get array j))^"\n")); 
            (printSurprisal tau p i (j+1) ppa array))
  |false -> (output_string p "\n");(flush p); array;;

(*The surprisal calculator driver. *)
let runSurprisal fcl ppa p i =  match fcl with
    [] -> (Array.make 1 0.0)
  |h::t -> match h with
	[] -> (Array.make 1 0.0)
      |hh::tt -> (printSurprisal hh.config.tau p i 0 ppa (calcSurp ppa (Array.make (Array.length ppa) 0.0) ((Array.length ppa)-1)));;    

(** The driver for the whole module, called from main.ml. *)
let surprisal fullConfigList p i= (runSurprisal fullConfigList (makeInput fullConfigList) p i);;


