(** Utility functions and types*)



(**{% \subsection{Data structures} %} *)

(**The data structure for words to be parsed.  I call them nodes.*)
type node = {word: string; lemma: string; cat: string; pos: int; parent: int; toPeek: bool};; 

(**Anything named dummy* is used for debugging and place-holding.*)
let dummyNode = {word="NULL";lemma="NULL";cat="NULL";pos=(-2);parent=(-2);toPeek=false};;

(**The data structure for the decay calculation for a word.  It just adds time information to a node.*)
type wordDecay = {tNode: node; timeSince: float list; fan: float;};;


(**h and d maps for the nivre configuration.  They hold the dependency structure that is being built.  Because this version is unlabeled, I really only use d, which, when given an integer position of a node, will return its head.  D should return the label of the dependency, and works that way if labels are ever used.*)
module H = Map.Make(Int32);;
module D  = Map.Make(Int32);;

(**The Nivre configuration is held in an item called {i config}.  It consists of: 
- a {i stack} of elements that have already been parsed, but may still need some dependents, 
- an input list {i tau} (both of these are lists of nodes), 
- {i h} and {i d}, the maps that hold the dependency information, 
- a list of nodes that have been parsed {i sofar} (this is just used for outputting, etc., to keep track of what might have been reduced off the stack), 
- a {i time} variable that tells you how much time has passed since parsing has started for this configuration (used for retrieval).  
- a list of nodes with retrieval information associated with them in {i decay}.*)
type config = {stack: (node list); tau: (node list); h: int H.t; d: string D.t; soFar: (node list); time: float; decay: wordDecay list};;

(**These are the possible transitions that can be taken from one configuration to another.*)
type operation = Shift | Reduce | Left | Right;;

(** This is the data structure for the transitions between configurations.  It contains:
- an {i operation}, 
- a string placeholder for the thing that's being acted on {i element}, 
- the {i prob}ability for the transition, the two nodes involved, 
- with the operation going from the {i fromWord} and to the {i toWord}, 
- the forward probability {i fprobProb} associated with the sequence of transitions that have led to this one,
- the lexical probability {i lexicalProb} associated with the transition (this is currently not used, only for a lexicalized parser)
- a boolean that is used to tell whether we have already looked at the lexical item beneath the part-of-speech (again, only used for a lexicalized parser).
*)
type arc = {op: operation; element: string; prob: float; fromWord: node; toWord: node; fprobProb: float; lexicalProb: float; peek: bool};;

(**This keeps track of the features that are used during parsing.*)
type model = Top | Next | Distance | Position | TopChildren | TopParent | TopLeft | TopRight | Stack1 | Stack3 | Stack2 | Stack1Next | Stack2Next | Stack3Next | StackSize;;

(**Because you can play with features being used together in a parse, we use a modelList to keep track of them.*)
type modelList = {list: (model list); name: string};;


(**This is the specification for the set that holds arcs, called ArcSet.*)
module type FLOAT = sig 
  type t 
  val compare : t -> t -> int
end 

module Arc : FLOAT with type t = arc = 
  struct 
    type t = arc
    let  compare (x:t) (y:t) = if (x.prob = y.prob) then 0 else if (x.prob > y.prob) then 1 else -1
  end 
module ArcSet = Set.Make(Arc) 

(*maybe used by surprisal?*)
(*let reverseLog x = 10.0**x;;*)

(**This keeps track of the forwardProbability associated with a configuration.
- {i fprob} is the forward probability itself
- {i zeros} is the number of zeroes that have been encountered for transition probabilities so far.  I don't think this is a problem anymore because of the new probability model, but just in case there are zeros, we can't multiply them across because it would make the forward probability be zero for the entire transition sequence.
- {i forSurprisalCalc} determines whether we should consider this to be the last forward probability associated with a word (before it's shifted onto the stack), in which case it should be the forward probability used in the surprisal calculation.
- {i lexProb} has something to do with lexicalized parsing, I can't remember what.*)
type fp = {fprob: float; zeros: int; forSurprisalCalc: bool; lexProb: float};;


(**Now we can finally define the data structure that holds the parser state, which I call an {i item}.  Original, I know.  Bad name, but I don't think it'll be changed anytime soon because it's so pervasive.  At any rate, it has all the information associated with a parser state.
- {i config} is the Nivre configuration.
- {i arcInfo} is the transition that is taken from this configuration to get to the next one.  It's an arc.
- {i history} lists all the past parser states that have led to this one.  In reverse order (so starting from the most recent).
- {i rank} is where this parser state places in relation to other parser states at this iteration, in terms of probabilities.  It's used to keep track of k-best parsing.
- {i forwardProb} is the forward probability associated with this parser state.  This is the probability that is used to rank the state.*)
type item = {config: config;
	     arcInfo: arc;
	     history: item list;
	       rank: int;
             forwardProb: fp};;


(** {% \subsection{Input from ocamlyacc} %}*)
(**The next three are used by depYacc to make the input sentence list.*)
let makeNode w l c p = {word=w; lemma=l; cat=c; pos=(-1); parent=p; toPeek=true};;                                                                                                 
let rec addPos wl i= match wl with
        [] ->[]
        |h::t -> {word=h.word;lemma=h.lemma;cat=h.cat;pos=i;parent=h.parent; toPeek=h.toPeek}::(addPos t (i+1))

let makeSentence wl= (addPos (List.rev wl) 0);;       


(** {% \subsection{Utility functions}\\
These are just functions that I use often across modules, so I put them here since they will be generated first.  They're not in any particular order. %} *)


(**Because I use maps in the Nivre configuration to hold the dependency information, it's easiest to use a fold to get the information out for outputs, etc.*)
let outputH h = H.fold (fun s n l -> (s, n)::l) h [];;                                                                                                   
let outputD d = D.fold (fun s n l -> (s, n)::l) d [];;      


(**Again for outputting, it's nice to have a string version of the transitions.*)
let string_of_op x = match x with
    Left -> "LEFT"
  |Right -> "RIGHT"
  |Shift -> "SHIFT"
  |Reduce -> "REDUCE";;

(**To calculate surprisal and retrieval, we use the natural log a lot.  This is just a shorthand. *)
let natural_log x = ((log x) /. (log 2.0));; 

(**These are used for inputting from ocamlyacc, I think.  They help to find out where there are problems in the file.*)
let elementType = ref ("pos");;
let linenum = ref 1;;  

(**This was used at one point, when we considered doing a generative model based on general POS tags. Might be useful for retrieval again. *)
(*let isVerbal x = match x with 
     "VBD"|"VBG"|"VBN"|"VBP"|"VBZ"|"VVFIN"|"VVIMP"|"VVINF"|"VVIZU"|"VVPP"|"VAFIN"|"VAIMP"|"VAINF"|"VAPP"|"VMFIN"|"VMINF"|"VMPP"|"VV"|"VA"|"VC"|"VE" -> true
  |_ -> false;;

let isNominal x = match x with
    "NN"|"NNS"|"NNP"|"NNPS"|"PRP"|"PRP$"|"NE"|"PDS"|"PDAT"|"PIS"|"PIAT"|"PPER"|"PPOSS"|"PPOSAT"|"NNE"|"NR"|"PN" -> true
  |_ -> false*)

(**The following were used when we played with different grains on the lexical items.  I won't get rid of them because they might still be useful. *)
type conditionElement = Category | Lexical | Lemma;;
type modelType = Generative | Conditional;;
let grain = ref Category;;
let conditionOn n = match !grain with
    Lexical -> n.word
    |Lemma -> n.lemma
    |_ -> n.cat
let genCondModel = ref Conditional;;


(*REMOVE*)
(*
    let stack1Model = {list=[Stack1];name="Stack1"};;                                                                                                                             
    let stack2Model = {list=[Stack2];name="Stack2"};;                                                                                                                             
    let stack3Model = {list=[Stack3];name="Stack3"};;                                                                                                                             
    let stack1NextModel = {list=[Stack1Next];name="Stack1Next"};;                                                                                                                 
    let stack2NextModel = {list=[Stack2Next];name="Stack2Next"};;                                                                                                                 
    let stack3NextModel = {list=[Stack3Next];name="Stack3Next"};;                                                                                                                
    let topModel = {list=[Top]; name="Top"};;                                                                                                                                     
    let nextModel = {list=[Next]; name="Next"};;                                                                                                                                  
    let distanceModel = {list = [Distance]; name="Distance"};;                                                                                                                    
    let positionModel = {list=[Position]; name="Position"};;                                                                                                                      
    let topChildrenModel = {list =[TopChildren]; name="TopChildren"};;                                                                                                            
    let topParentModel = {list = [TopParent]; name="TopParent"};;                                                                                                                 
    let topLeftModel = {list=[TopLeft]; name="TopLeft"};;                                                                                                                         
    let topRightModel = {list =[TopRight]; name="TopRight"};;                                                                                                                     
    let stackSizeModel = {list=[StackSize];name="StackSize"};;                                                                                                                    
    let stack3TopCModel = {list=[Stack3;TopChildren];name="Stack3TopChildren"};;                                                                                                  
    let allModel = {list=[Stack3;Stack1;Top;Next;Distance;Position;TopChildren;TopParent;TopLeft;TopRight;StackSize]; name="All"};;                                               
    let stacksModel = {list=[Stack3;Stack2;Stack1];name="Stacks"};;                                                                                                               
    let stacksNextModel = {list=[Stack3Next;Stack2Next;Stack1Next];name="StacksNext"};;                                                                                           
(*This list specifies the features that will actually be run.  Currently we're only running the stack features because they give us the best p and r*)
    let bestList = [stack3Model;stack3TopCModel;allModel;stack1Model];;                                                                                                           
    let theBest = [stacksModel;stack3Model;stack2Model;stack1Model];;                                                                                                             
*)
(*/REMOVE*)
